@@ -79,6 +79,7 @@ module Data.Bytes.Parser
 
     -- * Repetition
   , replicate
+  , listUntilEoi
 
     -- * Subparsing
   , delimit
@@ -104,23 +105,30 @@ module Data.Bytes.Parser
     -- of monadic @>>=@ can be helpful. If @C#@, @I#@, etc. never
     -- get used in your original source code, GHC will not introduce them.
   , bindFromCharToLifted
+  , bindFromCharToByteArrayIntInt
+  , bindFromWordToByteArrayIntInt
   , bindFromLiftedToIntPair
   , bindFromLiftedToInt
   , bindFromIntToIntPair
   , bindFromCharToIntPair
+  , bindFromLiftedToByteArrayIntInt
+  , bindFromByteArrayIntIntToLifted
   , bindFromMaybeCharToIntPair
   , bindFromMaybeCharToLifted
 
     -- * Specialized Pure
   , pureIntPair
+  , pureByteArrayIntInt
 
     -- * Specialized Fail
   , failIntPair
+  , failByteArrayIntInt
   ) where
 
 import Prelude hiding (any, fail, length, replicate, take, takeWhile)
 
 import Data.Bytes.Parser.Internal (Parser (..), Result#, ST#, boxBytes, fail, unboxBytes, uneffectful, uneffectful#, uneffectfulInt#)
+import Data.Bytes.Parser.Internal (failByteArrayIntInt)
 import Data.Bytes.Parser.Types (Result (Failure, Success), Slice (Slice))
 import Data.Bytes.Parser.Unsafe (cursor, expose, unconsume)
 import Data.Bytes.Types (Bytes (..), BytesN (BytesN))
@@ -128,6 +136,7 @@ import Data.Primitive (ByteArray (..))
 import Data.Primitive.Contiguous (Contiguous, Element)
 import Foreign.C.String (CString)
 import GHC.Exts (Char#, Int (I#), Int#, Word#, runRW#, (+#), (-#), (>=#))
+import GHC.Exts (ByteArray#)
 import GHC.ST (ST (..))
 import GHC.Word (Word32 (W32#), Word8)
 
@@ -135,6 +144,7 @@ import qualified Arithmetic.Nat as Nat
 import qualified Arithmetic.Types as Arithmetic
 import qualified Data.Bytes as B
 import qualified Data.Bytes.Parser.Internal as Internal
+import qualified Data.List as List
 import qualified Data.Primitive as PM
 import qualified Data.Primitive.Contiguous as C
 import qualified GHC.Exts as Exts
@@ -249,17 +259,13 @@ cstring :: e -> CString -> Parser e s ()
 cstring e (Exts.Ptr ptr0) = Parser
   ( \(# arr, off0, len0 #) s ->
     let go !ptr !off !len = case
-#if MIN_VERSION_base(4,16,0)
                  Exts.word8ToWord#
-#endif
             (Exts.indexWord8OffAddr# ptr 0#) of
           0## -> (# s, (# | (# (), off, len #) #) #)
           c -> case len of
             0# -> (# s, (# e | #) #)
             _ -> case Exts.eqWord# c (
-#if MIN_VERSION_base(4,16,0)
                  Exts.word8ToWord#
-#endif
               (Exts.indexWord8Array# arr off)) of
               1# -> go (Exts.plusAddr# ptr 1# ) (off +# 1# ) (len -# 1# )
               _ -> (# s, (# e | #) #)
@@ -519,15 +525,15 @@ peekRemaining = uneffectful $ \b@(Bytes _ off len) ->
 -- | Skip while the predicate is matched. This is always inlined.
 skipWhile :: (Word8 -> Bool) -> Parser e s ()
 {-# INLINE skipWhile #-}
-skipWhile f = go
+skipWhile f = goSkipWhile
  where
-  go =
+  goSkipWhile =
     isEndOfInput >>= \case
       True -> pure ()
       False -> do
         w <- anyUnsafe
         if f w
-          then go
+          then goSkipWhile
           else unconsume 1
 
 {- | The parser @satisfy p@ succeeds for any byte for which the
@@ -584,9 +590,7 @@ unboxWord32 (Parser f) = Parser
     (# s1, r #) -> case r of
       (# e | #) -> (# s1, (# e | #) #)
       (# | (# W32# a, b, c #) #) -> (# s1, (# | (#
-#if MIN_VERSION_base(4,16,0)
         Exts.word32ToWord#
-#endif
         a, b, c #) #) #)
   )
 {- FOURMOLU_ENABLE -}
@@ -612,9 +616,7 @@ boxWord32 (Parser f) = Parser
     (# s1, r #) -> case r of
       (# e | #) -> (# s1, (# e | #) #)
       (# | (# a, b, c #) #) -> (# s1, (# | (# W32# (
-#if MIN_VERSION_base(4,16,0)
         Exts.wordToWord32#
-#endif
         a), b, c #) #) #)
   )
 {- FOURMOLU_ENABLE -}
@@ -696,6 +698,28 @@ bindFromCharToLifted (Parser f) g =
             runParser (g y) (# arr, b, c #) s1
     )
 
+bindFromWordToByteArrayIntInt :: Parser s e Word# -> (Word# -> Parser s e (# ByteArray#, Int#, Int# #)) -> Parser s e (# ByteArray#, Int#, Int# #)
+{-# INLINE bindFromWordToByteArrayIntInt #-}
+bindFromWordToByteArrayIntInt (Parser f) g =
+  Parser
+    ( \x@(# arr, _, _ #) s0 -> case f x s0 of
+        (# s1, r0 #) -> case r0 of
+          (# e | #) -> (# s1, (# e | #) #)
+          (# | (# y, b, c #) #) ->
+            runParser (g y) (# arr, b, c #) s1
+    )
+
+bindFromCharToByteArrayIntInt :: Parser s e Char# -> (Char# -> Parser s e (# ByteArray#, Int#, Int# #)) -> Parser s e (# ByteArray#, Int#, Int# #)
+{-# INLINE bindFromCharToByteArrayIntInt #-}
+bindFromCharToByteArrayIntInt (Parser f) g =
+  Parser
+    ( \x@(# arr, _, _ #) s0 -> case f x s0 of
+        (# s1, r0 #) -> case r0 of
+          (# e | #) -> (# s1, (# e | #) #)
+          (# | (# y, b, c #) #) ->
+            runParser (g y) (# arr, b, c #) s1
+    )
+
 bindFromCharToIntPair :: Parser s e Char# -> (Char# -> Parser s e (# Int#, Int# #)) -> Parser s e (# Int#, Int# #)
 {-# INLINE bindFromCharToIntPair #-}
 bindFromCharToIntPair (Parser f) g =
@@ -718,6 +742,34 @@ bindFromLiftedToInt (Parser f) g =
             runParser (g y) (# arr, b, c #) s1
     )
 
+bindFromByteArrayIntIntToLifted ::
+     Parser s e (# ByteArray#, Int#, Int# #)
+  -> ((# ByteArray#, Int#, Int# #) -> Parser s e a)
+  -> Parser s e a
+{-# INLINE bindFromByteArrayIntIntToLifted #-}
+bindFromByteArrayIntIntToLifted (Parser f) g =
+  Parser
+    ( \x@(# arr, _, _ #) s0 -> case f x s0 of
+        (# s1, r0 #) -> case r0 of
+          (# e | #) -> (# s1, (# e | #) #)
+          (# | (# y, b, c #) #) ->
+            runParser (g y) (# arr, b, c #) s1
+    )
+
+bindFromLiftedToByteArrayIntInt ::
+     Parser s e a
+  -> (a -> Parser s e (# ByteArray#, Int#, Int# #))
+  -> Parser s e (# ByteArray#, Int#, Int# #)
+{-# INLINE bindFromLiftedToByteArrayIntInt #-}
+bindFromLiftedToByteArrayIntInt (Parser f) g =
+  Parser
+    ( \x@(# arr, _, _ #) s0 -> case f x s0 of
+        (# s1, r0 #) -> case r0 of
+          (# e | #) -> (# s1, (# e | #) #)
+          (# | (# y, b, c #) #) ->
+            runParser (g y) (# arr, b, c #) s1
+    )
+
 bindFromLiftedToIntPair :: Parser s e a -> (a -> Parser s e (# Int#, Int# #)) -> Parser s e (# Int#, Int# #)
 {-# INLINE bindFromLiftedToIntPair #-}
 bindFromLiftedToIntPair (Parser f) g =
@@ -776,6 +828,14 @@ pureIntPair a =
   Parser
     (\(# _, b, c #) s -> (# s, (# | (# a, b, c #) #) #))
 
+pureByteArrayIntInt ::
+  (# ByteArray#, Int#, Int# #) ->
+  Parser s e (# ByteArray#, Int#, Int# #)
+{-# INLINE pureByteArrayIntInt #-}
+pureByteArrayIntInt a =
+  Parser
+    (\(# _, b, c #) s -> (# s, (# | (# a, b, c #) #) #))
+
 failIntPair :: e -> Parser e s (# Int#, Int# #)
 {-# INLINE failIntPair #-}
 failIntPair e =
@@ -848,6 +908,20 @@ delimit esz eleftovers (I# n) (Parser f) =
         _ -> (# s0, (# esz | #) #)
     )
 
+{- | Apply the parser repeatedly until there is no more input left
+to consume. Collects the results into a list.
+-}
+listUntilEoi ::
+     Parser e s a -- ^ Parser to repeatedly apply until input is exhausted
+  -> Parser e s [a]
+listUntilEoi p = go []
+  where
+  go !acc = isEndOfInput >>= \case
+    True -> pure $! List.reverse acc
+    False -> do
+      a <- p
+      go (a : acc)
+
 {- | Replicate a parser @n@ times, writing the results into
 an array of length @n@. For @Array@ and @SmallArray@, this
 is lazy in the elements, so be sure the they result of the
