@@ -99,6 +99,7 @@ module Data.Bytes.Parser.Latin
     -- *** Fixed Length
   , hexFixedWord8
   , hexFixedWord16
+  , hexFixedWord16#
   , hexFixedWord32
   , hexFixedWord64
   , hexFixedWord128
@@ -701,11 +702,7 @@ upcastWord64Result :: Result# e Word# -> Result# e Word64
 {-# inline upcastWord64Result #-}
 upcastWord64Result (# e | #) = (# e | #)
 upcastWord64Result (# | (# a, b, c #) #) = (# | (# W64# (
-#if MIN_VERSION_base(4,17,0)
     Exts.wordToWord64# a
-#else
-    a
-#endif
     ), b, c #) #)
 {- FOURMOLU_ENABLE -}
 
@@ -774,9 +771,7 @@ upcastWord16Result :: Result# e Word# -> Result# e Word16
 {-# inline upcastWord16Result #-}
 upcastWord16Result (# e | #) = (# e | #)
 upcastWord16Result (# | (# a, b, c #) #) = (# | (# W16# (
-#if MIN_VERSION_base(4,16,0)
   Exts.wordToWord16#
-#endif
   a), b, c #) #)
 
 -- Precondition: the word is small enough
@@ -784,9 +779,7 @@ upcastWord32Result :: Result# e Word# -> Result# e Word32
 {-# inline upcastWord32Result #-}
 upcastWord32Result (# e | #) = (# e | #)
 upcastWord32Result (# | (# a, b, c #) #) = (# | (# W32# (
-#if MIN_VERSION_base(4,16,0)
   Exts.wordToWord32#
-#endif
   a), b, c #) #)
 
 -- Precondition: the word is small enough
@@ -794,9 +787,7 @@ upcastWord8Result :: Result# e Word# -> Result# e Word8
 {-# inline upcastWord8Result #-}
 upcastWord8Result (# e | #) = (# e | #)
 upcastWord8Result (# | (# a, b, c #) #) = (# | (# W8# (
-#if MIN_VERSION_base(4,16,0)
   Exts.wordToWord8#
-#endif
   a), b, c #) #)
 {- FOURMOLU_ENABLE -}
 
@@ -1188,9 +1179,7 @@ hexFixedWord32 e = Parser
     (# s1, r #) -> case r of
       (# err | #) -> (# s1, (# err | #) #)
       (# | (# a, b, c #) #) -> (# s1, (# | (# W32# (
-#if MIN_VERSION_base(4,16,0)
         Exts.wordToWord32#
-#endif
         a), b, c #) #) #)
   )
 {- FOURMOLU_ENABLE -}
@@ -1239,12 +1228,7 @@ hexFixedWord64 e = Parser
     (# s1, r #) -> case r of
       (# err | #) -> (# s1, (# err | #) #)
       (# | (# a, b, c #) #) -> (# s1, (# | (# W64# (
-#if MIN_VERSION_base(4,17,0)
           Exts.wordToWord64# a
-#else
-
-          a
-#endif
           ), b, c #) #) #)
   )
 {- FOURMOLU_ENABLE -}
@@ -1298,12 +1282,15 @@ hexFixedWord16 e = Parser
     (# s1, r #) -> case r of
       (# err | #) -> (# s1, (# err | #) #)
       (# | (# a, b, c #) #) -> (# s1, (# | (# W16# (
-#if MIN_VERSION_base(4,16,0)
         Exts.wordToWord16#
-#endif
         a), b, c #) #) #)
   )
 
+-- | Variant of hexFixedWord16 that returns an unboxed result. The result is
+-- a machine-sized word instead of a 16-bit word, but this function guarantees
+-- that only the low 16 bits may be non-zero. This is helpful because the
+-- result needs to be constructed as a machine-sized word. GHC's primitives
+-- for working with 16-bit words lower to most ISAs poorly.
 hexFixedWord16# :: e -> Parser e s Word#
 {-# noinline hexFixedWord16# #-}
 hexFixedWord16# e = uneffectfulWord# $ \chunk -> if length chunk >= 4
@@ -1314,9 +1301,9 @@ hexFixedWord16# e = uneffectfulWord# $ \chunk -> if length chunk >= 4
         !w3@(W# n3) = oneHex $ PM.indexByteArray (array chunk) (offset chunk + 3)
      in if | w0 .|. w1 .|. w2 .|. w3 /= maxBound ->
              (# |
-                (# (n0 `Exts.timesWord#` 4096##) `Exts.plusWord#`
-                   (n1 `Exts.timesWord#` 256##) `Exts.plusWord#`
-                   (n2 `Exts.timesWord#` 16##) `Exts.plusWord#`
+                (# (n0 `Exts.shiftL#` 12#) `Exts.plusWord#`
+                   (n1 `Exts.shiftL#` 8#) `Exts.plusWord#`
+                   (n2 `Exts.shiftL#` 4#) `Exts.plusWord#`
                    n3
                 ,  unI (offset chunk) +# 4#
                 ,  unI (length chunk) -# 4# #) #)
@@ -1334,9 +1321,7 @@ hexFixedWord8 e = Parser
     (# s1, r #) -> case r of
       (# err | #) -> (# s1, (# err | #) #)
       (# | (# a, b, c #) #) -> (# s1, (# | (# W8# (
-#if MIN_VERSION_base(4,16,0)
         Exts.wordToWord8#
-#endif
         a), b, c #) #) #)
   )
 {- FOURMOLU_ENABLE -}
